#include <xlic/licensing.h>

#include <xlic/client.hpp>
#include <xtl/exception.hpp>
#include <memory>

size_t sizes(
	const xlic::license_list & licenses,
	size_t & structs_size
) noexcept {
	structs_size = sizeof(xlic_license_list);
	size_t strings_size = 0;
	for (const auto & pair : licenses) {
		const auto & license = pair.second;
		structs_size += sizeof(xlic_license);
		strings_size += 29; // base64-encoded length of license key + null byte
		strings_size += license.name.length() + 1;
		strings_size += license.description.length() + 1;
		for (const auto & pair : license.attributes) {
			structs_size += sizeof(xlic_license_attribute);
			strings_size += pair.first.length() + 1; // attribute name
			strings_size += pair.second.length() + 1; // attribute value
		}
	}
	return structs_size + strings_size;
}

template<typename T>
inline T * pointer(void * ptr, size_t offset) noexcept {
	typedef char * byte_pointer;
	typedef T * pointer_type;
	return pointer_type(byte_pointer(ptr) + offset);
}

xlic_license_state copy(const xlic::license_state state) noexcept {
	switch (state) {
		case xlic::license_state::inactive: return XLIC_STATE_INACTIVE;
		case xlic::license_state::active: return XLIC_STATE_ACTIVE;
		case xlic::license_state::grace: return XLIC_STATE_GRACE;
		case xlic::license_state::invalid: return XLIC_STATE_INVALID;
		case xlic::license_state::unknown: return XLIC_STATE_UNKNOWN;
		case xlic::license_state::force_invalid: return XLIC_STATE_INVALID;
	}
	return XLIC_STATE_UNKNOWN;
}

xlic_license_type copy(const xlic::license_type type) noexcept {
	switch (type) {
		case xlic::license_type::physical_license: return XLIC_TYPE_PHYSICAL;
		case xlic::license_type::virtual_license: return XLIC_TYPE_VIRTUAL;
		case xlic::license_type::unknown: return XLIC_TYPE_UNKNOWN;
	}
	return XLIC_TYPE_UNKNOWN;
}

char * copy(
	const std::string & src,
	char * & strings
) {
	const auto size = src.length() + 1;
	auto result = static_cast<char *>(memcpy(strings, src.c_str(), size));
	strings += size;
	return result;
}

char * copy(
	const xlic::license_key & key,
	char * & strings
) {
	return copy(to_string(key), strings);
}

xlic_license_attribute * copy(
	const xlic::attribute_list & src,
	xlic_license_attribute * & attributes,
	char * & strings
) {
	xlic_license_attribute * first_attribute = attributes;
	auto attribute = first_attribute;
	for (const auto & pair : src) {
		attribute->name = copy(pair.first, strings);
		attribute->value = copy(pair.second, strings);
		attributes++;
	}
	return first_attribute;
}

xlic_license * copy(
	const xlic::license_list & src,
	xlic_license * licenses,
	xlic_license_attribute * attributes,
	char * strings
) {
	auto dst = licenses;
	for (const auto & pair : src) {
		const auto & license = pair.second;
		dst->state = copy(license.state);
		dst->type = copy(license.type);
		dst->key = copy(license.license_key, strings);
		dst->name = copy(license.name, strings);
		dst->description = copy(license.description, strings);
		dst->valid_after = license.valid_after;
		dst->valid_for = license.valid_for;
		dst->grace_days = license.grace_days;
		dst->attribute_count = license.attributes.count();
		dst->attributes = copy(license.attributes, attributes, strings);
		dst++;
	}
	return licenses;
}

void copy(
	const xlic::license_list & src,
	void * ptr,
	size_t structs_size
) {
	auto list = static_cast<xlic_license_list *>(ptr);
	auto licenses = pointer<xlic_license>(list, sizeof(xlic_license_list));
	auto attributes = pointer<xlic_license_attribute>(licenses, sizeof(xlic_license) * src.count());
	auto strings = pointer<char>(ptr, structs_size);
	list->count = src.count();
	list->licenses = copy(src, licenses, attributes, strings);
}

extern "C" int xlic_licenses(const char * socket_filename, xlic_license_list ** result) {
	try {
		const auto licenses = xlic::client(socket_filename).list_licenses();
		size_t structs_size = 0;
		if (void * ptr = malloc(sizes(licenses, structs_size))) {
			copy(licenses, ptr, structs_size);
			*result = static_cast<xlic_license_list *>(ptr);
			return 0;
		}
	} catch (const xlic::client_exception &) {
		return XLIC_RESULT_DAEMON_ERROR;
	} catch (const xtl::unix_socket_exception &) {
		return XLIC_RESULT_DAEMON_ERROR;
	} catch (const xlic::result_code & code) {
		return int(code);
	} catch (...) {}
	return -1;
}
