#include <xlic/request_license.hpp>

#include <xtl/exception.hpp>
#include <xtl/https_stream.hpp>
#include <xlic/json.hpp>
#include <xlic/worker.hpp>

#if defined(DEBUG_REQUEST_LICENSE) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("request_license[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("request_license[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xlic {

request_license::request_license(
	xlic::session * session,
	xlic::worker & worker,
	xlic::request_id request_id,
	const std::string & name,
	const std::string & username,
	const std::string & password
)
: task(session, worker, request_id)
, m_result_code(xlic::result_code::unknown)
, m_thread([this, name, username, password] () {
	if (m_worker.foreign_license_file()) {
		m_result_code = xlic::result_code::foreign;
	} else {
		try {
			const xlic::license_name xlic_name(name);
			const xlic::username xlic_username(username);
			const xlic::password xlic_password(password);
			xtl::https_stream stream(XLIC_SERVER_HOSTNAME);
			stream.post("/create");
			stream.useragent(XLIC_USERAGENT);
			stream.content_type(XLIC_CONTENT_TYPE);
			stream << xlic::json_encode(
				m_worker.serial_number(),
				m_worker.mac_addresses(),
				m_worker.system_name(),
				xlic_name,
				xlic_username,
				xlic_password
			);
			stream.flush();
			const xtl::json js(stream);
			try {
				const auto serial_number = xlic::serial_number(js["serial_number"].string());
				const auto public_key = from_base64<xtl::rsa_public_key>(js["public_key"].string());
				const auto license = from_json<xlic::license>(js["license"]);
				const auto system_name = xlic::system_name(js["system_name"].string());
				m_worker.add(serial_number, system_name, public_key, license);
				m_result_code = xlic::result_code::success;
			} catch (const xtl::json_exception &) {
				m_result_code = xlic::result_code::generic_error;
				const auto message = js["message"].string();
				if (message == "rejected") {
					m_result_code = xlic::result_code::rejected;
				}
			}
		} catch (const std::exception & e) {
			fdebug("%s\n", e.what());
			m_result_code = xlic::result_code::generic_error;
		}
	}
	m_worker.completed(this);
}) {}

void request_license::write_result(xtl::ostream & stream) {
	if (m_result_code == xlic::result_code::success) {
		stream << xlic::packet_size(xlic::packet_header_size);
		stream << xlic::packet_type::success;
		stream << m_request_id;
	} else {
		stream << xlic::packet_size(xlic::packet_header_size + sizeof(m_result_code));
		stream << xlic::packet_type::error;
		stream << m_request_id;
		stream << m_result_code;
	}
}

} // namespace xlic
