#if	USE_ROCKEY

#include <xlic/rockey.hpp>
#include <xlic/rockey.h>
#include <list>
#include <stdexcept>
#include <cstring>
#include <array>
#include <zlib.h>

struct info_raw {
	uint32_t version;
	uint32_t hwid;
	uint64_t timestamp;
	uint16_t magic;
	uint16_t serial_length;
	char serial[16];
} __attribute__((packed));

inline constexpr uint16_t xor_hash(
	const uint16_t a,
	const uint16_t b,
	const uint16_t c,
	const uint16_t d,
	const uint16_t e,
	const uint16_t f,
	const uint16_t g
) noexcept {
	return ((((((((((a + g) ^ b) + g) ^ c) + g) ^ d) + g) ^ e) + g) ^ f) + g;
}

inline uint32_t hash_block(
	const void * ptr,
	const size_t size
) noexcept {
	typedef const Bytef * pointer;
	return crc32(0, pointer(ptr), size);
}

inline uint32_t hash_serial(const std::string & serial) noexcept {
	return hash_block(serial.data(), serial.length());
}

template<typename T, size_t size>
inline void random_fill(std::array<T, size> & buffer) noexcept {
	for (size_t i = 0; i < size; ++i) {
		buffer[i] = rand();
	}
}

inline void calc_xor(
	void * target,
	const void * mask,
	const size_t target_size,
	const size_t mask_size
) noexcept {
	auto t = static_cast<uint8_t *>(target);
	auto m = static_cast<const uint8_t *>(mask);
	for (size_t i = 0; i < target_size; ++i) {
		t[i] ^= m[i % mask_size];
	}
}

template<typename T, size_t s1, typename U, size_t s2>
inline void calc_xor(
	std::array<T, s1> & target,
	const std::array<U, s2> & mask
) noexcept {
	calc_xor(target.data(), mask.data(), sizeof(T) * s1, sizeof(U) * s2);
}

std::vector<rockey> rockey::enumerate(
	uint16_t p1,
	uint16_t p2,
	uint16_t p3,
	uint16_t p4
) {
	std::list<rockey> dongles;
	uint16_t handle = 0;
	uint32_t hwid = 0;
	// RY_FIND performs password test for "reasons" and has built-in nanosleep({tv_sec=1, tv_nsec=0}) :(
	auto result = Rockey(RY_FIND, &handle, &hwid, nullptr, &p1, &p2, &p3, &p4, nullptr);
	while (result == ERR_SUCCESS) {
		dongles.emplace_back(handle, hwid);
		result = Rockey(RY_FIND_NEXT, &handle, &hwid, nullptr, &p1, &p2, &p3, &p4, nullptr);
	}
	if (result != ERR_NOMORE) {
		throw rockey_exception(result);
	}
	return std::vector<rockey> {
		std::make_move_iterator(std::begin(dongles)),
		std::make_move_iterator(std::end(dongles))
	};
}

void rockey::open(
	uint16_t p1,
	uint16_t p2,
	uint16_t p3,
	uint16_t p4
) {
	uint32_t unknown = 0; // will be set to 16 after RY_OPEN for reasons
	const auto result = Rockey(RY_OPEN, &m_handle, &m_hwid, &unknown, &p1, &p2, &p3, &p4, nullptr);
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
	m_opened = true;
}

void rockey::close() noexcept {
	// Dongles cannot be reopened after closing them, must renumerate
	Rockey(RY_CLOSE, &m_handle, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr);
	m_opened = false;
}

void rockey::read_memory(
	void * buffer,
	uint16_t size,
	uint16_t offset
) {
	if (size == 0) {
		return;
	} else if (size + offset > 1000) {
		throw std::out_of_range("Memory area is only 1000 bytes long");
	}
	typedef uint8_t * pointer;
	const auto result = Rockey(RY_READ, &m_handle, nullptr, nullptr, &offset, &size, nullptr, nullptr, pointer(buffer));
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
}

void rockey::write_memory(
	const void * buffer,
	uint16_t size,
	uint16_t offset
) {
	if (size == 0) {
		return;
	} else if (size + offset > 1000) {
		throw std::out_of_range("Memory area is only 1000 bytes long");
	}
	typedef uint8_t * pointer;
	const auto result = Rockey(RY_WRITE, &m_handle, nullptr, nullptr, &offset, &size, nullptr, nullptr, pointer(buffer));
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
}

void rockey::set_module(
	uint16_t index,
	uint16_t value
) {
	if (index >= 64) {
		throw std::out_of_range("Up to 64 modules can be defined");
	}
	uint16_t can_decrement = false;
	const auto result = Rockey(RY_SET_MODULE, &m_handle, nullptr, nullptr, &index, &value, &can_decrement, nullptr, nullptr);
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
}

void rockey::set_algo(
	uint16_t index,
	const std::string & algorithm
) {
	if (index >= 128) {
		throw std::out_of_range("Up to 128 instructions can be programmed");
	}
	std::vector<uint8_t> buffer(algorithm.length() + 1);
	buffer.insert(buffer.begin(), std::begin(algorithm), std::end(algorithm));
	// The ALU behaves a bit weird and documentation is vague, I only figured this out after looking at one of
	// the screenshots. There are invisible start/end markers defined when programming a set of instructions.
	// I think the start marker is used to determine valid offsets when calling RY_CALCULATE. The ALU appears to
	// stop executing when it hits the end marker, thus allowing more than one algorithm to be defined.
	const auto result = Rockey(RY_WRITE_ARITHMETIC, &m_handle, nullptr, nullptr, &index, nullptr, nullptr, nullptr, buffer.data());
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
}

inline void do_algo(
	uint16_t func,
	uint16_t handle,
	uint32_t index,
	uint32_t arg,
	uint16_t & a,
	uint16_t & b,
	uint16_t & c,
	uint16_t & d
) {
	const auto result = Rockey(func, &handle, &index, &arg, &a, &b, &c, &d, nullptr);
	if (result != ERR_SUCCESS) {
		throw rockey_exception(result);
	}
}

void rockey::call_algo(
	uint16_t index,
	algo_mode mode,
	uint16_t & a,
	uint16_t & b,
	uint16_t & c,
	uint16_t & d,
	uint32_t arg
) {
	switch (mode) {
		case algo_mode::A: return do_algo(RY_CALCULATE1, m_handle, index, arg, a, b, c, d);
		case algo_mode::B: return do_algo(RY_CALCULATE2, m_handle, index, arg, a, b, c, d);
		case algo_mode::C: return do_algo(RY_CALCULATE3, m_handle, index, arg, a, b, c, d);
	}
	throw std::out_of_range("Invalid mode");
}

rockey::info rockey::read() {
	std::array<uint16_t, 250> buffer;
	read_memory(buffer.data(), 500, 500);
	for (size_t i = 0; i < 250; i += 4) {
		call_algo(
			0,
			algo_mode::C,
			buffer[i],
			buffer[i + 1],
			buffer[i + 2],
			buffer[i + 3],
			i % 64
		);
	}
	info_raw raw;
	memcpy(&raw, buffer.data(), sizeof(raw));
	return rockey::info{
		be32toh(raw.version),
		be32toh(raw.hwid),
		be64toh(raw.timestamp),
		be16toh(raw.magic),
		std::string(raw.serial, std::min(size_t(be16toh(raw.serial_length)), sizeof(info_raw::serial)))
	};
}

bool rockey::is_xorcom_dongle(
	const std::string & serial,
	uint16_t magic
) {
	const uint32_t hash1 = hash_block(&m_hwid, sizeof(m_hwid));
	const uint32_t hash2 = hash_serial(serial);
	uint16_t a = hash1 & 0xffff, b = hash1 >> 16, c = hash2 & 0xffff, d = hash2 >> 16;
	const uint16_t e = m_hwid >> 16, f = m_hwid & 0xffff, g = magic;
	const auto r1 = xor_hash(a, b, c, d, e, f, g), r2 = xor_hash(b, c, d, e, f, r1, g);
	const auto r3 = xor_hash(c, d, e, f, r1, r2, g), r4 = xor_hash(d, e, f, r1, r2, r3, g);
	call_algo(4, algo_mode::A, a, b, c, d, m_hwid & 0x3f);
	return ((a ^ r1) | (b ^ r2) | (c ^ r3) | (d ^ r4)) == 0;
}

void rockey::program(const std::string & serial) {
	set_algo(0, "A=A^E,B=B^F,C=C^G,D=D^H"); // Calculate XOR of m[i0..i3] and P1..P4
	set_algo(4, // Calculate custom hash of P1..P4 HWID and m[HWID & 3f]
		"H=A+G,H=H^B,H=H+G,H=H^C,H=H+G,H=H^D,H=H+G,H=H^E,H=H+G,H=H^F,A=H+G,"
		"H=B+G,H=H^C,H=H+G,H=H^D,H=H+G,H=H^E,H=H+G,H=H^F,H=H+G,H=H^A,B=H+G,"
		"H=C+G,H=H^D,H=H+G,H=H^E,H=H+G,H=H^F,H=H+G,H=H^A,H=H+G,H=H^B,C=H+G,"
		"H=D+G,H=H^E,H=H+G,H=H^F,H=H+G,H=H^A,H=H+G,H=H^B,H=H+G,H=H^C,D=H+G"
	);
	std::array<uint16_t, 64> modules;
	random_fill(modules);
	for (size_t i = 0; i < 64; ++i) {
		set_module(i, modules[i]);
	}
	std::array<uint16_t, 250> buffer;
	random_fill(buffer);
	info_raw raw;
	raw.hwid = htobe32(m_hwid);
	raw.version = htobe32(1);
	raw.timestamp = htobe64(time(nullptr));
	raw.magic = htobe16(modules[m_hwid & 0x3f]);
	const auto serial_length = std::min(serial.length(), sizeof(info_raw::serial));
	raw.serial_length = htobe16(serial_length);
	serial.copy(raw.serial, serial_length);
	memcpy(buffer.data(), &raw, sizeof(raw));
	calc_xor(buffer, modules);
	write_memory(buffer.data(), 500, 500);
}

const char * rockey_exception::what() const noexcept {
	static char unknown[] = "Unknown rockey error (00000)";
	switch (m_code) {
		case ERR_SUCCESS: return "ERR_SUCCESS";
		case ERR_NO_ROCKEY: return "ERR_NO_ROCKEY";
		case ERR_INVALID_PASSWORD: return "ERR_INVALID_PASSWORD";
		case ERR_INVALID_PASSWORD_OR_ID: return "ERR_INVALID_PASSWORD_OR_ID";
		case ERR_SETID: return "ERR_SETID";
		case ERR_INVALID_ADDR_OR_SIZE: return "ERR_INVALID_ADDR_OR_SIZE";
		case ERR_UNKNOWN_COMMAND: return "ERR_UNKNOWN_COMMAND";
		case ERR_NOTBELEVEL3: return "ERR_NOTBELEVEL3";
		case ERR_READ: return "ERR_READ";
		case ERR_WRITE: return "ERR_WRITE";
		case ERR_RANDOM: return "ERR_RANDOM";
		case ERR_SEED: return "ERR_SEED";
		case ERR_CALCULATE: return "ERR_CALCULATE";
		case ERR_NO_OPEN: return "ERR_NO_OPEN";
		case ERR_OPEN_OVERFLOW: return "ERR_OPEN_OVERFLOW";
		case ERR_NOMORE: return "ERR_NOMORE";
		case ERR_NEED_FIND: return "ERR_NEED_FIND";
		case ERR_DECREASE: return "ERR_DECREASE";
		case ERR_AR_BADCOMMAND: return "ERR_AR_BADCOMMAND";
		case ERR_AR_UNKNOWN_OPCODE: return "ERR_AR_UNKNOWN_OPCODE";
		case ERR_AR_WRONGBEGIN: return "ERR_AR_WRONGBEGIN";
		case ERR_AR_WRONG_END: return "ERR_AR_WRONG_END";
		case ERR_AR_VALUEOVERFLOW: return "ERR_AR_VALUEOVERFLOW";
		case ERR_TOOMUCHTHREAD: return "ERR_TOOMUCHTHREAD";
		case ERR_INVALID_RY4S: return "ERR_INVALID_RY4S";
		case ERR_INVALID_PARAMETER: return "ERR_INVALID_PARAMETER";
		case ERR_INVALID_TIMEVALUE: return "ERR_INVALID_TIMEVALUE";
		case ERR_SET_DES_KEY: return "ERR_SET_DES_KEY";
		case ERR_DES_ENCRYPT: return "ERR_DES_ENCRYPT";
		case ERR_DES_DECRYPT: return "ERR_DES_DECRYPT";
		case ERR_SET_RSAKEY_N: return "ERR_SET_RSAKEY_N";
		case ERR_SET_RSAKEY_D: return "ERR_SET_RSAKEY_D";
		case ERR_RSA_ENCRYPT: return "ERR_RSA_ENCRYPT";
		case ERR_RSA_DECRYPT: return "ERR_RSA_DECRYPT";
		case ERR_INVALID_LENGTH: return "ERR_INVALID_LENGTH";
		case ERR_RECEIVE_NULL: return "ERR_RECEIVE_NULL";
		case ERR_UNKNOWN_SYSTEM: return "ERR_UNKNOWN_SYSTEM";
		case ERROR_UNINIT_TIME_UNIT: return "ERROR_UNINIT_TIME_UNIT";
		case ERR_UNKNOWN: return "ERR_UNKNOWN";
		case ERR_XUTEST: return "ERR_XUTEST";
	}
	snprintf(unknown, sizeof(unknown), "Unknown rockey error (%u)", m_code);
	return unknown;
}

#endif // __x86_64__
